use std::fs::{self, read_to_string};

use proc_macro::TokenStream;
use quote::{format_ident, quote};
use syn::{parse, token::Comma, LitBool, Result};
use syn::{
    parse::{Parse, ParseStream},
    spanned::Spanned,
    Error, Ident, Lit, LitInt,
};

const DATA_DIR: &str = "data";
const EXAMPLE_DIR: &str = "examples";
const VAR_NAME: &str = "DATA";

fn get_text<T: Spanned>(item: T) -> Result<String> {
    let span = item.span();
    match span.source_text() {
        Some(text) => Ok(text),
        None => {
            let start = span.start();
            let end = span.end();
            Err(Error::new(span, format!("Failed to get sourcetext for {}:{}-{}:{}", start.line, start.column, end.line, end.column,)))
        },
    }
}

#[derive(Debug, Clone, Default)]
struct IncludeData {
    var: String,
    year: String,
    day: String,
    example: bool,
    parts: usize,
}

impl Parse for IncludeData {
    fn parse(input: ParseStream) -> Result<Self> {
        let mut v = Self::default();
        v.var = if input.peek(Ident) && !input.peek(LitInt) {
            let t = get_text(input.parse::<Ident>().unwrap())?;
            input.parse::<Comma>()?;
            t
        } else {
            VAR_NAME.to_string()
        };
        v.year = get_text(input.parse::<LitInt>()?)?;
        input.parse::<Comma>()?;
        v.day = get_text(input.parse::<Lit>()?)?;
        if input.peek2(LitBool) {
            input.parse::<Comma>()?;
            v.example = input.parse::<LitBool>()?.value;
        }
        if input.peek2(LitInt) {
            input.parse::<Comma>()?;
            v.parts = input.parse::<LitInt>()?.base10_parse()?;
        }
        if v.parts == 0 {
            v.parts = 1;
        }
        Ok(v)
    }
}

fn fallback(input: IncludeData, message: String) -> TokenStream {
    if input.parts != 1 {
        let postfixes = (1..=input.parts).map(|i| format_ident!("{}_{}", input.var, i));
        quote! {
            #[doc = #message]
            #(
                const #postfixes: &str = "";
            )*
        }
        .into()
    } else {
        let name = format_ident!("{}", input.var);
        quote! {
            #[doc = #message]
            const #name: &str = "";
        }
        .into()
    }
}

/// Macro to include data and examples.
/// The First Argument(Optional) is the name/prefix for the constant,
/// The Second is the Year
/// The Third is the day/filename without the .txt at the end
/// the fourth is if its an example or not
/// and the last is the amount of parts in the file, if it is greater than 1 the first argument is treated as an Prefix.
#[proc_macro]
pub fn include_aoc(data: TokenStream) -> TokenStream {
    let input = match parse::<IncludeData>(data) {
        Ok(data) => data,
        Err(error) => return error.into_compile_error().into(),
    };
    let mut content = Default::default();
    if !input.example {
        let path = format!("{}/{}/{}.txt", DATA_DIR, input.year, input.day);
        match fs::exists(path.clone()) {
            Ok(true) => content = read_to_string(path).unwrap(),
            Ok(false) => {},
            Err(error) => return fallback(input, format!("Failed to read file: {}", error)),
        };
    }
    if content.is_empty() {
        let path = format!("{}/{}/{}.txt", EXAMPLE_DIR, input.year, input.day);
        content = match fs::exists(path.clone()) {
            Ok(true) => read_to_string(path).unwrap(),
            Ok(false) => Default::default(),
            Err(error) => return fallback(input, format!("Failed to read file: {}", error)),
        };
    }
    if input.parts != 1 {
        let parts = content.splitn(input.parts, "\n\n");
        let suffix = (1..=input.parts).map(|i| format_ident!("{}_{}", input.var, i));
        quote! {
            #(
                const #suffix: &str = #parts;
            )*
        }
    } else {
        let name = format_ident!("{}", input.var);
        quote! {
            const #name: &str = #content;
        }
    }
    .into()
}
