use std::fmt::Display;

use crate::enum_alias;

use super::{Kartesian, KartesianIterator};
use num::*;

#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Default)]
pub enum KartesianDirection {
    #[default]
    None,
    TopLeft,
    Top,
    TopRight,
    Left,
    Right,
    BottomLeft,
    Bottom,
    BottomRight,
}

impl KartesianDirection {
    enum_alias!(KartesianDirection, NorthWest, TopLeft);
    enum_alias!(KartesianDirection, North, Top);
    enum_alias!(KartesianDirection, Up, Top);
    enum_alias!(KartesianDirection, NorthEast, TopRight);
    enum_alias!(KartesianDirection, West, Left);
    enum_alias!(KartesianDirection, Center, None);
    enum_alias!(KartesianDirection, East, Right);
    enum_alias!(KartesianDirection, SouthWest, BottomLeft);
    enum_alias!(KartesianDirection, South, Bottom);
    enum_alias!(KartesianDirection, Down, Bottom);
    enum_alias!(KartesianDirection, SouthEast, BottomLeft);
}

impl KartesianDirection {
    pub fn vector<T: Integer>(self) -> Kartesian<T> {
        Kartesian {
            x: match self {
                Self::Left | Self::None | Self::Right => T::zero(),
                _ => T::one(),
            },
            y: match self {
                Self::Top | Self::None | Self::Bottom => T::zero(),
                _ => T::one(),
            },
        }
    }

    #[inline]
    pub fn vector_abs<T: Integer + From<u8>>(self) -> Kartesian<T> {
        self.vector()
    }

    pub fn clockwise(self, diagonal: bool) -> Self {
        match (self, diagonal) {
            (Self::TopLeft, _) => Self::Top,
            (Self::Top, true) => Self::TopRight,
            (Self::Top, false) => Self::Right,
            (Self::TopRight, _) => Self::Right,
            (Self::Right, true) => Self::BottomRight,
            (Self::Right, false) => Self::Bottom,
            (Self::BottomRight, _) => Self::Bottom,
            (Self::Bottom, true) => Self::BottomLeft,
            (Self::Bottom, false) => Self::Left,
            (Self::BottomLeft, _) => Self::Left,
            (Self::Left, true) => Self::TopLeft,
            (Self::Left, false) => Self::Top,
            (Self::None, _) => Self::None,
        }
    }

    pub fn anticlockwise(self, diagonal: bool) -> Self {
        match (self, diagonal) {
            (Self::TopLeft, _) => Self::Left,
            (Self::Left, true) => Self::BottomLeft,
            (Self::Left, false) => Self::Bottom,
            (Self::BottomLeft, _) => Self::Bottom,
            (Self::Bottom, true) => Self::BottomRight,
            (Self::Bottom, false) => Self::Right,
            (Self::BottomRight, _) => Self::Right,
            (Self::Right, true) => Self::TopRight,
            (Self::Right, false) => Self::Top,
            (Self::TopRight, _) => Self::Top,
            (Self::Top, true) => Self::TopLeft,
            (Self::Top, false) => Self::Left,
            (Self::None, _) => Self::None,
        }
    }

    pub fn up(self) -> Self {
        match self {
            Self::Top | Self::TopLeft | Self::TopRight => self,
            Self::Left => Self::TopLeft,
            Self::None => Self::Top,
            Self::Right => Self::TopRight,
            Self::BottomLeft => Self::Left,
            Self::Bottom => Self::None,
            Self::BottomRight => Self::Right,
        }
    }
    pub fn down(self) -> Self {
        match self {
            Self::Bottom | Self::BottomLeft | Self::BottomRight => self,
            Self::TopLeft => Self::Left,
            Self::Top => Self::None,
            Self::TopRight => Self::Right,
            Self::Left => Self::BottomLeft,
            Self::None => Self::Bottom,
            Self::Right => Self::BottomRight,
        }
    }
    pub fn left(self) -> Self {
        match self {
            Self::TopLeft | Self::Left | Self::BottomLeft => self,
            Self::Top => Self::TopLeft,
            Self::None => Self::Left,
            Self::Bottom => Self::BottomLeft,
            Self::TopRight => Self::Top,
            Self::Right => Self::None,
            Self::BottomRight => Self::Bottom,
        }
    }
    pub fn right(self) -> Self {
        match self {
            Self::TopRight | Self::Right | Self::BottomRight => self,
            Self::Top => Self::TopRight,
            Self::None => Self::Right,
            Self::Bottom => Self::BottomRight,
            Self::TopLeft => Self::Top,
            Self::Left => Self::None,
            Self::BottomLeft => Self::Bottom,
        }
    }
    pub fn neg(self) -> Self {
        match self {
            Self::TopLeft => Self::BottomRight,
            Self::Top => Self::Bottom,
            Self::TopRight => Self::BottomLeft,
            Self::Left => Self::Right,
            Self::None => Self::None,
            Self::Right => Self::Left,
            Self::BottomLeft => Self::TopRight,
            Self::Bottom => Self::Top,
            Self::BottomRight => Self::TopLeft,
        }
    }

    pub fn iter(none: bool, diagonal: bool, straight: bool) -> KartesianIterator {
        KartesianIterator {
            i: 0,
            none,
            diagonal,
            straight,
        }
    }
}

impl Display for KartesianDirection {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let _ = match *self {
            Self::Top | Self::TopLeft | Self::TopRight => f.write_str("top"),
            Self::Bottom | Self::BottomLeft | Self::BottomRight => f.write_str("bottom"),
            Self::None => return f.write_str("none"),
            _ => std::fmt::Result::Ok(()),
        };
        let _ = match *self {
            Self::TopLeft | Self::Left | Self::BottomLeft => f.write_str("left"),
            Self::TopRight | Self::Right | Self::BottomRight => f.write_str("right"),
            _ => std::fmt::Result::Ok(()),
        };
        std::fmt::Result::Ok(())
    }
}
